/*
FILE  main.c

DOC   This program simulates a three-dimensional Lennard-Jones fluid.

LANG  C plus MPI message passing.

HIS   1) Parallel version originally written by University of
HIS      Southampton.
HIS   2) Adopted to run on the Intel iPSC/2 computer at Daresbury
HIS      Laboratory, and placed in the public domain through the
HIS      CCP5 programme.
HIS   3) Adopted to run on the Intel Paragon and enhanced by David W.
HIS      Walker at Oak Ridge National Laboratory, Tennessess, USA
HIS   4) Converted to use MPI message passing library by David W. Walker
HIS      at University of Wales Cardiff in July 1997.
HIS   5) Converted from Fortran to C in November 1997.

*/

#ifndef SEQ
#include <mpi.h>
#endif

#include "moldyn.h"
#include <stdio.h>
#include <stdlib.h>
#include <omp.h>

float sigma;
float rcut;
float dt;
float eqtemp;
float dens;
float boxlx;
float boxly;
float boxlz;
float sfx;
float sfy;
float sfz;
float cellix;
float celliy;
float celliz;
float sr6;
float vrcut;
float dvrcut;
float dvrc12;
float freex;
float vg;
float wg;
float kg;
float tmpx;
float *scrat;
float posvel[INREAL];
int   nprosx;
int   nprosy;
int   nprosz;
int   nstep;
int   nequil;
int   isave;
int   iscale;
int   nstart;
int   isvunf;
int   isvpfs;
int   nc;
int   natoms;
int   mx;
int   my;
int   mz;
int   istart;
int   where;
int   outfst=0;
int   jfirst=0;
int   outfrm;
int   map[3][3][3];

int comm_size;
int numthreads;

float ace;
float acv;
float ack;
float acp;
float acesq;
float acvsq;
float acksq;
float acpsq;
int iprint;

// Timing variables - MJC
#ifdef TIMING
   double tstart, moveastart, movebstart, sumstart, hloopstart, startupstart, iterstart, forcesstart, movoutstart;
   double tend, moveaend, movebend, sumend, hloopend, startupend, iterend, forcesend, movoutend;
   double ttime, moveatime, movebtime, sumtime, hlooptime, startuptime, itertime, forcestime, movouttime;
#endif

int main (int argc, char* argv[])
{


   int rank;
   int xi, yi, zi;
   float xo, yo, zo;
   /* dynamically allocate arrays */
   float *rx = (float *) malloc(sizeof(float) * NMAX);
   float *ry = (float *) malloc(sizeof(float) * NMAX);
   float *rz = (float *) malloc(sizeof(float) * NMAX);
   float *vx = (float *) malloc(sizeof(float) * NMAX2);
   float *vy = (float *) malloc(sizeof(float) * NMAX2);
   float *vz = (float *) malloc(sizeof(float) * NMAX2);
   int *head = (int *) malloc(sizeof(int) * NCELL);
   int *list = (int *) malloc(sizeof(int) * NMAX);
   scrat = (float *) malloc(sizeof(float) * SCRSIZ);
   int i;

   int   natm=0;
   int ierror, input_parameters();
   int jstart, step, itemp;
   float potential, virial, kinetic;
   int icell;
#ifndef SEQ
   MPI_Init(&argc, &argv);
#endif
#ifndef SEQ
   MPI_Comm_rank(MPI_COMM_WORLD, &rank);
   MPI_Comm_size(MPI_COMM_WORLD, &comm_size);
#include "moldyn.h"
#else
   rank = RANKHOST;
   comm_size = 1;
#endif

   #ifdef TIMING
      if(I_AM_HOST) {
         tstart = MPI_Wtime();
      }
   #endif
   #ifdef TIMING
      if(I_AM_HOST) {
         startupstart = MPI_Wtime();
      }
   #endif
   #pragma omp parallel shared(numthreads)
   {
      numthreads = omp_get_num_threads();
   }

   ierror = input_parameters (rank, &xi, &yi, &zi, &xo, &yo, &zo);

   if (ierror==0) {
      if (nstart==1 || nstart==2)
         printf ("\n Sorry, this option is not available");
      else if (nstart==3)
         cold_start (rank, rx, ry, rz, vx, vy, vz, xi, yi, zi, xo, yo, zo,&natm);

   }

   loop_initialise (rank);

   if (where==1) {
     movout (rx, ry, rz, vx, vy, vz, head, list, xi, yi, zi, &natm);
     force (&potential, &virial, rx, ry, rz, head, list, natm);
   }

   jstart = (istart==-1) ? 1 : istart + where;
   #ifdef TIMING
      if(I_AM_HOST) {
         startupend = MPI_Wtime();
         startuptime += (startupend-startupstart);
      }
   #endif
   #ifdef TIMING
      if(I_AM_HOST) {
         iterstart = MPI_Wtime();
      }
   #endif

   for(step=jstart;step<=nstep;step++) {

      #ifdef TIMING
         if(I_AM_HOST) {
            moveastart = MPI_Wtime();
         }
      #endif
      if ((where==1)||(step>jstart)) movea (rx, ry, rz, vx, vy, vz, natm);
      #ifdef TIMING
         if(I_AM_HOST) {
            moveaend = MPI_Wtime();
            moveatime += (moveaend-moveastart);
         }
      #endif
      #ifdef TIMING
      if(I_AM_HOST)
      {
         movoutstart = MPI_Wtime();
      }
      #endif
      movout (rx, ry, rz, vx, vy, vz, head, list, xi, yi, zi, &natm);
      #ifdef TIMING
      if(I_AM_HOST)
      {
         movoutend = MPI_Wtime();
         movouttime += (movoutend - movoutstart);
      }
      #endif
      itemp = abs(isave);
      outfrm = 0;
      if (step%itemp == 0) outfrm++;
      itemp = abs(isvunf);
      if (step%itemp ==0) outfrm += 10;
      if (outfrm>0) outcon (step, 0, rank, rx, ry, rz, vx, vy, vz,
                            xo, yo, zo, natm);
      #ifdef TIMING
         if(I_AM_HOST) {
            forcesstart = MPI_Wtime();
         }
      #endif
      force (&potential, &virial, rx, ry, rz, head, list, natm);
      #ifdef TIMING
         if(I_AM_HOST) {
            forcesend = MPI_Wtime();
            forcestime += (forcesend - forcesstart);
         }
      #endif
      #ifdef TIMING
         if(I_AM_HOST) {
            movebstart = MPI_Wtime();
         }
      #endif
      moveb (&kinetic, vx, vy, vz, natm);
      #ifdef TIMING
         if(I_AM_HOST) {
            movebend = MPI_Wtime();
            movebtime += (movebend-movebstart);
         }
      #endif
      #ifdef TIMING
         if(I_AM_HOST) {
            sumstart = MPI_Wtime();
         }
      #endif
      sum_energies (potential, kinetic, virial);
      #ifdef TIMING
         if(I_AM_HOST) {
            sumend = MPI_Wtime();
            sumtime += (sumend-sumstart);
         }
      #endif
      #ifdef TIMING
         if(I_AM_HOST) {
            hloopstart = MPI_Wtime();
         }
      #endif
      hloop (kinetic, step, rank, vx, vy, vz, natm);
      #ifdef TIMING
         if(I_AM_HOST) {
            hloopend = MPI_Wtime();
            hlooptime += (hloopend-hloopstart);
         }
      #endif
   }

   #ifdef TIMING
   if(I_AM_HOST)
   {
      iterend = MPI_Wtime();
      itertime += (iterend - iterstart);
   }
   #endif

   outfrm = 0;
   if (isave > 0) outfrm++;
   if (isvunf > 0) outfrm += 10;

   if (isave >0 || isvunf > 0)
      outcon (nstep, 1, rank, rx, ry, rz, vx, vy, vz, xo, yo, zo, natm);

   #ifdef TIMING
      if(I_AM_HOST) {
         tend = MPI_Wtime();
         ttime += (tend-tstart);
      }
   #endif


   if (I_AM_HOST)
   {
      tidyup ();
   }

   free(rx);
   free(ry);
   free(rz);
   free(vx);
   free(vy);
   free(vz);
   free(list);
   free(head);
   free(scrat);

#ifndef SEQ
   MPI_Finalize ();
#endif
}
